from string import ascii_uppercase, digits
from typing import Any

from shortuuid.django_fields import ShortUUIDField


class ShortIdField(ShortUUIDField):
    def __init__(self, *args: Any, **kwargs: Any):
        kwargs["length"] = kwargs.get("length", 11)
        kwargs["max_length"] = kwargs.get("max_length", 20)
        kwargs["alphabet"] = kwargs.get("alphabet", ascii_uppercase + digits)
        super().__init__(*args, **kwargs)



import hashlib
import json
from contextlib import suppress
from datetime import datetime
from itertools import filterfalse, tee, zip_longest
from operator import methodcaller
from string import ascii_uppercase, digits
from types import SimpleNamespace
from typing import Any, Iterable, List, Literal

from django.db.models import QuerySet
from shortuuid import ShortUUID


class unset:
    pass


def identity(value: any):
    return value


def safe_int(val: Any, default=0, min_=None, max_=None) -> int:
    with suppress(Exception):
        val = int(val)
        if min_ is not None:
            val = max(val, min_)
        if max_ is not None:
            val = min(val, max_)
        return int(val)
    return default


def queryset_iterator(queryset: QuerySet, chunk_size: int = 1000):
    total = queryset.count()
    for start in range(0, total, chunk_size):
        end = min(start + chunk_size, total)
        yield queryset[start:end], start, end


def extract(
    params: dict, *keys, default: any = unset, how: Literal["get", "pop"] = "get"
) -> tuple:
    assert how in ["get", "pop"], f"how must be one of ['get', 'pop'], got {how}."

    if default is not unset:
        if not isinstance(default, (list, tuple, set)):
            default = [default] * len(keys)
        pairs = zip_longest(keys, default, fillvalue=None)
        return tuple(methodcaller(how, key, def_)(params) for key, def_ in pairs)
    return tuple(methodcaller(how, key)(params) for key in keys)


def deduplicate(
    seq: Iterable, key: callable = identity, keep: Literal["first", "last"] = "first"
) -> List:
    assert keep in [
        "first",
        "last",
    ], f"keep must be one of ['first', 'last'], got {keep}."

    seen = {}
    for item in seq:
        k = key(item)
        if keep == "first" and k not in seen:
            seen[k] = item
        if keep == "last":
            seen[k] = item

    return list(seen.values())


def to_timestamp(dt, default=0, round_ts=False) -> int | float:
    if isinstance(dt, datetime):
        if round_ts:
            return round(dt.timestamp() * 1e3)
        return dt.timestamp()
    return default


def partition(iterable: Iterable, pred: callable = identity):
    t1, t2 = tee(iterable)
    return filter(pred, t1), filterfalse(pred, t2)


def hexer(s: str, alg: Literal["md5", "sha256"]) -> str:
    mapping = {"md5": hashlib.md5, "sha256": hashlib.sha256}
    hasher_alg = mapping.get(alg)
    if s and alg and hasher_alg:
        hasher = hasher_alg()
        hasher.update(s.encode("utf-8"))
        return hasher.hexdigest()
    return s


def objectify(*dicts: dict, default=None) -> SimpleNamespace | list[SimpleNamespace]:
    def convert(data: dict):
        return json.loads(json.dumps(data), object_hook=lambda d: SimpleNamespace(**d))

    objects = [convert(d) for d in dicts]
    if len(objects) == 1:
        return next(iter(objects), default)
    return objects


def shortid(prefix: str, length: int = 10):
    return prefix + ShortUUID(alphabet=ascii_uppercase + digits).random(length=length)



from django.db import models
from django.utils import timezone

from core.models.manager import BaseManager, BaseQuerySet


class TimeStampedModel(models.Model):
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True


class DeletedQuerySet(BaseQuerySet):
    def delete(self, deleter_id: str = ""):
        for obj in self:
            obj.delete(deleter_id, save=False)
        return self.bulk_update(self, ["deleted", "deleter_id"])

    def destroy(self):
        for obj in self:
            obj.destroy()


class DeletedManager(BaseManager):
    def get_queryset(self):
        return DeletedQuerySet(self.model, using=self._db).filter(deleted__isnull=True)

    def include_deleted(self):
        return DeletedQuerySet(self.model, using=self._db)


class DeletedModel(models.Model):
    deleted = models.DateTimeField(null=True)
    deleter_id = models.CharField(max_length=20, default="")

    class Meta:
        abstract = True

    objects = DeletedManager()

    def delete(self, deleter_id: str = "", save: bool = True):
        if not self.deleted:
            self.deleted = timezone.now()
            self.deleter_id = deleter_id or self.deleter_id
            save and self.save()

    def restore(self, save: bool = True):
        if self.deleted:
            self.deleted = None
            self.deleter_id = ""
            save and self.save()

    def destroy(self, using=None, keep_parents=False):
        super().delete(using=using, keep_parents=keep_parents)


class HistoryModel(models.Model):
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)
    creator_id = models.CharField(max_length=20, default="")
    updater_id = models.CharField(max_length=20, default="")

    class Meta:
        abstract = True



from django.conf import settings
from rest_framework.permissions import BasePermission


class IsInternal(BasePermission):
    def has_permission(self, request, view):
        internal_key = request.META.get("HTTP_X_INTERNAL")
        if internal_key:
            return internal_key == settings.INTERNAL_KEY
        return False



from typing import Literal

from rest_framework import serializers
from rest_framework.serializers import ModelSerializer, Serializer

from core.utils import utils

ExtraHow = Literal["get", "pop"]


class BaseSerializer(Serializer):
    def pop(self, key):
        validated_data: dict = self.validated_data
        return validated_data.pop(key)

    def extract(
        self, *fields: str, how: ExtraHow = "pop", default=utils.unset
    ) -> tuple:
        fields = fields or [f for f in self.fields]
        return utils.extract(self.validated_data, *fields, how=how, default=default)


class BaseModelSerializer(ModelSerializer, BaseSerializer):
    def get_fields(self):
        fields = super().get_fields()
        if exclude_fields := self.context.get("exclude_fields", []):
            fields = {
                key: value for key, value in fields.items() if key not in exclude_fields
            }
        return fields


class IdSerializer(BaseSerializer):
    id = serializers.CharField(required=True, allow_blank=False)



import jwt
from django.utils.functional import SimpleLazyObject
from rest_framework.authentication import get_authorization_header
from rest_framework.exceptions import AuthenticationFailed
from rest_framework.request import Request


class TokenData:
    __slots__ = ("auth", "team", "user")

    def __init__(self, data: dict):
        self.auth = data.pop("aid")
        self.team = data.get("tid")
        self.user = data.get("uid")


def get_token_data(request: Request):
    try:
        auth = get_authorization_header(request).split()
        access = auth[1].decode()
        access_data = jwt.decode(jwt=access, options={"verify_signature": False})
        token = TokenData(access_data)
        return token
    except Exception as ex:
        raise AuthenticationFailed() from ex


class JwtTokenMiddleware:
    def __init__(self, get_response: callable):
        self.get_response = get_response

    def __call__(self, request: Request):
        request.token = SimpleLazyObject(lambda: get_token_data(request))
        response = self.get_response(request)
        return response



from rest_framework import status
from rest_framework.renderers import JSONRenderer


class ResponseRenderer(JSONRenderer):
    def render(self, data, accepted_media_type=None, renderer_context=None):
        omit_view_sets = {"Websocket"}

        if renderer_context:
            view = renderer_context.get("view")
            response = renderer_context.get("response")
            if response and view.basename not in omit_view_sets:
                if response.data is None:
                    response.data = data = {}
                if isinstance(data, dict):
                    if status.is_success(response.status_code):
                        data.setdefault("ok", True)
                    if status.is_client_error(response.status_code):
                        data.setdefault("ok", False)
                    if status.is_server_error(response.status_code):
                        data.setdefault("ok", False)
        return super().render(data, accepted_media_type, renderer_context)



import pytest
from django.conf import settings
from faker import Factory, Faker
from rest_framework.test import APIClient
from types import SimpleNamespace

factory = Factory.create()


@pytest.fixture()
def base_client():
    api_client = APIClient()
    return api_client


@pytest.fixture()
def client():
    api_client = APIClient()
    access_token = settings.ACCESS_TOKEN
    api_client.credentials(HTTP_AUTHORIZATION=f"Token {access_token}")
    api_client.token = SimpleNamespace(team="T0123456789", user="U0123456789")  # hardcoded
    return api_client


@pytest.fixture()
def internal_client():
    api_client = APIClient()
    internal_key = settings.INTERNAL_KEY
    api_client.credentials(HTTP_X_INTERNAL=internal_key)
    return api_client


@pytest.fixture()
def fake():
    return Faker()



import pytest
from faker import Factory, Generator
from rest_framework.test import APIClient

from core.utils import utils


class TestBase:
    @staticmethod
    def objectify(data: dict):
        return utils.objectify(data)


@pytest.mark.django_db(databases=["default", "replication"])
class ApiTestBase(TestBase):
    fake: Generator
    client: APIClient

    @pytest.fixture(autouse=True)
    def setup(self, client):
        self.client = client
        self.fake = Factory.create()

    def client_request(
        self,
        url: str,
        *,
        method: str = None,
        data: dict = None,
        format: str = None,
        ok: bool = None,
        status: int = None,
        client: APIClient = None
    ):
        data = data or {}
        format = format or "json"
        method = method or "post"
        assert method in ("get", "post", "put", "patch", "delete")
        client = client or self.client
        client_method = getattr(client, method)
        response = client_method(url, data=data, format=format)
        if status is not None:
            assert response.status_code == status
        resp = self.objectify(response.data)
        if ok is not None:
            assert resp.ok == ok
        return resp


@pytest.mark.django_db(databases=["default", "replication"])
class UnitTestBase(TestBase):
    pass



from datetime import datetime

from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from rest_framework import serializers


class TimestampField(serializers.DateTimeField):
    default_error_messages = {
        "invalid": _(
            "Timestamp has wrong format. Use one of these formats instead: {format}."
        ),
    }

    def to_representation(self, value: datetime):
        if not value:
            return None
        if self.context.get("round_ts", False):
            return round(value.timestamp() * 1e3)
        return value.timestamp()

    def to_internal_value(self, value):
        if not value:
            return None

        if isinstance(value, datetime):
            return timezone.make_naive(value, timezone.utc)

        try:
            value = float(value)
        except (TypeError, ValueError):
            self.fail("invalid")

        dt_value = None
        for val in (value, value / 1e3):
            try:
                dt_value = datetime.utcfromtimestamp(val)
            except (TypeError, ValueError):
                pass
            else:
                break

        if dt_value is None:
            self.fail("invalid")

        if timezone.is_aware(dt_value):
            return timezone.make_naive(dt_value, timezone.utc)
        return dt_value



from copy import deepcopy

from django.core.exceptions import PermissionDenied
from django.http import Http404
from rest_framework import exceptions
from rest_framework.response import Response
from rest_framework.views import set_rollback

from core.exceptions import ApiException


def exception_handler(exc, context):
    if isinstance(exc, Http404):
        exc = exceptions.NotFound()
    elif isinstance(exc, PermissionDenied):
        exc = exceptions.PermissionDenied()

    if isinstance(exc, exceptions.APIException):
        headers = {}
        if getattr(exc, "auth_header", None):
            headers["WWW-Authenticate"] = exc.auth_header  # noqa
        if getattr(exc, "wait", None):
            headers["Retry-After"] = "%d" % exc.wait  # noqa

        if isinstance(exc, ApiException):
            data = exc.detail
        else:
            codes = exc.get_codes()
            if isinstance(codes, dict):
                errors = [
                    {"field": key, "error": next(iter(values))}
                    for key, values in codes.items()
                ]
            elif isinstance(codes, list):
                errors = [{"error": code} for code in codes]
            else:
                errors = [{"error": codes}]
            data = deepcopy(next(iter(errors), {}))
            if len(errors) > 1:
                data.update(errors=errors)

        set_rollback()
        return Response(data, status=exc.status_code, headers=headers)

    return None



class Const:
    def __setattr__(self, name: str, value):
        raise ValueError("Cannot change %s.%s." % self.__class__.__name__, name)


class CommonConst(Const):
    OFF = 0
    ON = 1
    BLANK = ""
    SPACE = " "
    NEW_LINE = "\n"
    NA = "N/A"
    ZERO = 0
    ONE = 1
    TWO = 2
    ACTIVE = 1
    DEACTIVATED = 0



from copy import deepcopy
from operator import itemgetter
from typing import List

from rest_framework import status
from rest_framework.exceptions import APIException

from core.utils import utils


def _get_error_details(errors: List[dict]) -> dict:
    error = deepcopy(next(iter(errors)))
    if len(errors) > 1:
        error.update(errors=errors)
    return error


class ApiException(APIException):
    status_code = status.HTTP_400_BAD_REQUEST

    def __init__(
        self,
        error: str = None,
        errors: List[dict] = None,
        code: int = None,
        **kwargs,
    ):
        if error:
            error = {"error": error, **kwargs}
            errors = [error, *errors] if errors else [error]
            errors = utils.deduplicate(errors, key=itemgetter("error"))
        if not errors:
            errors = [{"error": "invalid"}]
        if code:
            self.status_code = code

        self.detail = _get_error_details(errors)



from urllib import parse

import requests
from django.conf import settings

from .base import BaseService


class ProxyService(BaseService):
    _scheme: str = "http"
    _ssl: bool = False
    _host: str
    _port: int

    @classmethod
    def _get_url(cls, path: str) -> str:
        scheme = cls._scheme
        if cls._ssl:
            scheme = f"{scheme}s"
        netloc = f"{cls._host}:{cls._port}"
        url = parse.urlunsplit((scheme, netloc, path, "", ""))
        return url

    @classmethod
    def _get_base_headers(cls, internal: bool = False) -> dict[str, str]:
        base_headers = {}
        if internal:
            base_headers.update({"X-Internal": settings.INTERNAL_KEY})
        return base_headers

    @classmethod
    def post(
        cls,
        path,
        data: dict,
        key: str = None,
        keys: list[str] = None,
        internal: bool = False,
        **kwargs,
    ) -> dict:
        data = data or {}
        url = cls._get_url(path)
        headers = kwargs.pop("headers", {})
        base_headers = cls._get_base_headers(internal=internal)
        response = requests.post(
            url=url, json=data, headers={**headers, **base_headers}, **kwargs
        )
        resp = response.json()
        if key:
            resp = resp.get(key)
        if keys:
            resp = {key: resp.get(key) for key in keys}
        return resp



from __future__ import annotations

from collections import UserDict
from typing import Literal


class DefaultDescriptor:
    def __init__(self, factory):
        self.factory = factory

    def __set_name__(self, owner, name):
        self.name = name

    def __get__(self, instance: Saga, owner: type[Saga]):
        if instance is None:
            return owner
        value = self.factory()
        setattr(instance, self.name, value)
        return value


class State(UserDict):
    pass


class Step:
    def action(self, state: State):
        pass

    def compensate(self, state: State):
        pass


class Saga:
    state: State = DefaultDescriptor(State)
    steps: list[Step] = DefaultDescriptor(list)
    status: Literal["success", "rollback", "error"]
    _execute_exc: Exception
    _rollback_exc: Exception
    _current_index: int
    _error_index: int

    def run(self):
        try:
            self._execute()
        except Exception as exc1:
            self._execute_exc = exc1
            try:
                self._rollback()
                self.status = "rollback"
            except Exception as exc2:
                self._rollback_exc = exc2
                self._error_index = self._current_index
                self.status = "error"
                self.on_error()
                raise exc2
            self.on_failure()
            raise exc1
        else:
            self.status = "success"
            self.on_success()

    def _execute(self):
        self._current_index = 0
        for idx, step in enumerate(self.steps):
            self._current_index = idx
            step.action(self.state)

    def _rollback(self):
        for idx in range(self._current_index - 1, -1, -1):
            self._current_index = idx
            self.steps[idx].compensate(self.state)

    def on_success(self):
        pass

    def on_failure(self):
        pass

    def on_error(self):
        pass

    @property
    def is_success(self):
        return self.status == "success"

    @property
    def is_rollback(self):
        return self.status == "rollback"

    @property
    def is_error(self):
        return self.status == "error"



class _Null:
    """See: http://en.wikipedia.org/wiki/Null_Object_pattern"""

    def __init__(self, _unicode="", _repr="Null"):
        self.__dict__["_unicode"] = _unicode
        self.__dict__["_str"] = _unicode
        self.__dict__["_repr"] = _repr

    def __unicode__(self):
        return self.__getattribute__("_unicode")

    def __str__(self):
        return self.__getattribute__("_str")

    def __repr__(self):
        return self.__getattribute__("_repr")

    def __nonzero__(self):
        return False

    def __len__(self):
        return 0

    def __getattr__(self, attr):
        if attr == "tolist":
            raise AttributeError
        return self

    def __setattr__(self, attr, value):
        pass

    def __getitem__(self, item):
        return self

    def __setitem__(self, item, value):
        pass

    def __delitem__(self, item):
        pass

    def __call__(self, *args, **kwargs):
        return self

    def __iter__(self):
        return self

    def next(self):
        raise StopIteration

    __next__ = next


class NullContext:
    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        pass


unset = _Null(_unicode="unset", _repr="unset")
Null = _Null()



from django.conf import settings

DEV_ENVIRONMENT = "dev"
TEST_ENVIRONMENT = "test"
STAGE_ENVIRONMENT = "stage"
PROD_ENVIRONMENT = "prod"


def is_dev(env: str = settings.APP_ENVIRONMENT) -> bool:
    return env == DEV_ENVIRONMENT


def is_test(env: str = settings.APP_ENVIRONMENT) -> bool:
    return env == TEST_ENVIRONMENT


def is_stage(env: str = settings.APP_ENVIRONMENT) -> bool:
    return env == STAGE_ENVIRONMENT


def is_prod(env: str = settings.APP_ENVIRONMENT) -> bool:
    return env == PROD_ENVIRONMENT



from functools import partial

from django.http import HttpResponse
from rest_framework import status
from rest_framework.decorators import action
from rest_framework.viewsets import ViewSet

from core.constants import Const


class BaseViewSet(ViewSet):
    pass


class HttpMethod(Const):
    GET = "get"
    POST = "post"
    PUT = "put"
    PATCH = "patch"
    DELETE = "delete"
    HEAD = "head"
    OPTIONS = "options"
    TRACE = "trace"


post = partial(action, methods=[HttpMethod.POST], detail=False)


def unauthorized(request):
    return HttpResponse("Unauthorized", status=status.HTTP_403_FORBIDDEN)



from __future__ import annotations

import json

from django.conf import settings
from pika import BasicProperties, BlockingConnection, ConnectionParameters, PlainCredentials
from pika.channel import Channel
from pika.connection import Connection


class ConnectionDescriptor:
    def __init__(self, host: str, port: int, username: str, password: str):
        self.host = host
        self.port = port
        self.username = username
        self.password = password

    def __set_name__(self, owner, name):
        self.name = name

    def __get__(self, instance, cls: type[BaseQueue]):
        if instance is None:
            assert all([self.host, self.port, self.username, self.password])

            credentials = PlainCredentials(username=self.username, password=self.password)
            parameters = ConnectionParameters(host=self.host, port=self.port, credentials=credentials)
            connection = BlockingConnection(parameters=parameters)
            setattr(cls, self.name, connection)
        raise RuntimeError("Cannot access class attributes from concrete instance.")


class ChannelDescriptor:
    def __set_name__(self, owner, name):
        self.name = name

    def __get__(self, instance, cls: type[BaseQueue]):
        if instance is None:
            channel = cls._connection.channel()
            setattr(cls, self.name, channel)
        raise RuntimeError("Cannot access class attributes from concrete instance.")


class BaseQueue:
    _connection: Connection = ConnectionDescriptor(
        host=getattr(settings, "RABBITMQ_BROKER_HOST", None),
        port=getattr(settings, "RABBITMQ_BROKER_PORT", None),
        username=getattr(settings, "RABBITMQ_BROKER_USERNAME", None),
        password=getattr(settings, "RABBITMQ_BROKER_PASSWORD", None),
    )
    _channel: Channel = ChannelDescriptor()
    _exchange: str

    @classmethod
    def publish(
        cls,
        data: dict,
        queue: str,
        exchange: str = "",
        mandatory: bool = False,
        properties: BasicProperties = None,
        **kwargs,
    ):
        data = data or {}
        exchange = exchange or cls._exchange
        body = json.dumps(data).encode("utf-8")
        properties = properties or BasicProperties(content_type="application/json", content_encoding="utf-8", **kwargs)

        cls._channel.basic_publish(
            body=body,
            exchange=exchange,
            routing_key=queue,
            properties=properties,
            mandatory=mandatory,
        )



from django.conf import settings
from core.services import ProxyService


class AuthService(ProxyService):
    _host = settings.MICROSERVICE.AUTH_HOST
    _port = settings.MICROSERVICE.AUTH_PORT


class TeamService(ProxyService):
    _host = settings.MICROSERVICE.TEAMS_HOST
    _port = settings.MICROSERVICE.TEAMS_PORT


class UserService(ProxyService):
    _host = settings.MICROSERVICE.USERS_HOST
    _port = settings.MICROSERVICE.USERS_PORT


class ConversationService(ProxyService):
    _host = settings.MICROSERVICE.CONVERSATIONS_HOST
    _port = settings.MICROSERVICE.CONVERSATIONS_PORT
